package es.us.dit.ejemplo.config;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.web.util.matcher.AntPathRequestMatcher;

 /* http://docs.spring.io/spring-security/site/docs/3.2.x/guides/hellomvc.html*/
 /* The name of the configureGlobal method is not important. However, 
 * it is important to only configure AuthenticationManagerBuilder 
 * in a class annotated with either @EnableWebSecurity, 
 * @EnableWebMvcSecurity, @EnableGlobalMethodSecurity, 
 * or @EnableGlobalAuthentication */
@EnableWebSecurity
public class SecurityConfig  extends WebSecurityConfigurerAdapter {

    @Autowired
    public void configureGlobal(AuthenticationManagerBuilder auth) throws Exception {
        auth
            .inMemoryAuthentication()
                .withUser("user").password("password").roles("USER")
                .and()
                .withUser("user2").password("1234").roles("USER", "ADMIN");
    }
    
    @Override
    protected void configure(HttpSecurity http) throws Exception {
    
    http
        // Have to disable it for POST methods:
        // http://stackoverflow.com/a/20608149/1199132
        .csrf().disable()

        // Logout and redirection:
        // http://stackoverflow.com/a/24987207/1199132
        .logout()
                .logoutRequestMatcher(new AntPathRequestMatcher("/logout"))
                .invalidateHttpSession(true)
                .logoutSuccessUrl("/login.xhtml")
                .and()
        .authorizeRequests()
                // Some filters enabling url regex:
                // http://stackoverflow.com/a/8911284/1199132
                .regexMatchers(
                        "\\A/page1.xhtml\\?param1=true\\Z",
                        "\\A/page2.xhtml.*")
                .permitAll()
                //Permit access for all to error and denied views
                .antMatchers("/error/**")
                .permitAll()
                // Only access with admin role
                .antMatchers("/config/**")
                .hasRole("ADMIN")
                //Permit access only for some roles
                .antMatchers("/person.xhtml")
                .hasAnyRole("ADMIN", "MANAGEMENT")
                //If user doesn't have permission, forward him to login page
                .and()
        .formLogin()
                .loginPage("/login.xhtml")
                .loginProcessingUrl("/login")
                .defaultSuccessUrl("/main.xhtml")
                .failureUrl("/error/denied.xhtml") // default is /login?error
                .and()
        .exceptionHandling().accessDeniedPage("/error/denied.xhtml");
    }
}