package es.us.dit.ejemplo.pictionary;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;

import org.primefaces.context.RequestContext;
import org.primefaces.push.EventBus;
import org.primefaces.push.EventBusFactory;
 
import javax.faces.application.FacesMessage;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;
 
@ManagedBean
@ViewScoped
public class PictionaryView implements Serializable {
     
    //private final PushContext pushContext = PushContextFactory.getDefault().getPushContext();
 
    private final EventBus eventBus = EventBusFactory.getDefault().eventBus();
 
    @ManagedProperty("#{pictionaryUsers}")
    private PictionaryUsers users;
    
    @ManagedProperty("#{pictionaryScores}")
    private PictionaryScores scores;
 
    private String keyword;
    
    private String winner;
    
    private boolean controlledDisconnection = false;
    
    private List<String> palabras = Arrays.asList("mesa", "deseo", "tallo", "amargo",
    		"sombrio", "rio", "lampara", "martillo", "musica", "blanco", "soñar", 
    		"sediento", "nausea", "bello", "amarillo", "blanco", "hombre", "ventana",
    		"pan", "plaza", "profundo", "aspero", "justicia", "mantequilla", "blanco",
    		"ciudadano", "muchacho", "doctor", "montaña", "araña", "salud", "ciruela",
    		"ladron", "casa", "alfiler", "biblia", "leon", "verde", "sal", "tijeras",
    		"negro", "rojo", "recuerdo", "alegria", "cordero", "dormir", "rebaño", "cama");
    
    private List<String> colores = Arrays.asList("#848484", "#FF00BF","#0080FF","#FFBF00",
    											"#FF8000","#FF0000", "#D7DF01", "#380B61");
    
    private String color = colores.get((int) Math.floor(Math.random()*(this.colores.size())));
    
    private String globalMessage;
    
    private String username;
     
    private boolean loggedIn;
     
    private String privateUser;
     
    private final static String CHANNEL = "/pictionaryroom/";
    
    public String getWinner() {
		return winner;
	}
    
    public void setWinner(String winner) {
		this.winner = winner;
	}
    
    public String getKeyword() {
		return keyword;
	}
    
    public void setKeyword(String keyword) {
		this.keyword = keyword;
	}
 
    public PictionaryUsers getUsers() {
        return users;
    }
 
    public void setUsers(PictionaryUsers users) {
        this.users = users;
    }
    
    public PictionaryScores getScores() {
        return scores;
    }
 
    public void setScores(PictionaryScores scores) {
        this.scores = scores;
    }
     
    public String getPrivateUser() {
        return privateUser;
    }
 
    public void setPrivateUser(String privateUser) {
        this.privateUser = privateUser;
    }
 
    public String getGlobalMessage() {
        return globalMessage;
    }
 
    public void setGlobalMessage(String globalMessage) {
        this.globalMessage = globalMessage;
    }
     
    public String getUsername() {
        return username;
    }
    public void setUsername(String username) {
        this.username = username;
    }
     
    public boolean isLoggedIn() {
        return loggedIn;
    }
    public void setLoggedIn(boolean loggedIn) {
        this.loggedIn = loggedIn;
    }
 
    public void sendGlobal() {
    	// uses the messageDecoder
    	eventBus.publish(CHANNEL + "*", username + ": " + globalMessage + ": " + color); 
        
    	// check the keyword
    	if (globalMessage.equalsIgnoreCase(keyword)) {
    		//push end information
            eventBus.publish(CHANNEL + "*", new Message(true, username, true));
    		
            // increase 10 points to the winner
            scores.increase(users.getPositionOfUser(username), 10);
            // show winner dialog and start timer to change painter
    		RequestContext requestContext = RequestContext.getCurrentInstance();
    		requestContext.execute("PF('dialogEndWinner').show()");
    		requestContext.execute("PF('timer').start()");
    		// stop game timer
    	}
    	
        globalMessage = null;
        
        
    }
    
    public void noWinners() {
    	RequestContext requestContext = RequestContext.getCurrentInstance();
    	//requestContext.execute("PF('timerGame').stop()");

    	//TODO WS no winners message
    	
    	if (users.size() == 1) {
    		//I'm the only player
    		requestContext.execute("PF('dlg').show()");
    	} else if (users.size() == 2) {
    		// There is another player
    		//TODO change painter
    		requestContext.execute("PF('noWinnersDialog').show()");
        	requestContext.execute("PF('timerRedirect').start()");
        	eventBus.publish(CHANNEL + users.getNextUser(this.username), new Message(true, true, false));
  
    	} else if (users.size() > 2) {
    		// There is more than 2 players
    		requestContext.execute("PF('noWinnersDialog').show()");
        	requestContext.execute("PF('timerRedirect').start()");
    		// next player is going to paint
        	String nextPlayer = users.getNextUser(this.username);
        	// send change painter message
        	eventBus.publish(CHANNEL + nextPlayer, new Message(true, true, false));
    		// Rest are going to wait for the next game
    		for (int i=0 ; i<users.size()-2 ; i++) {
    			// update next player
    			nextPlayer = users.getNextUser(nextPlayer);
    			// send wait message
    			eventBus.publish(CHANNEL + nextPlayer, new Message(true, true));
    		}
    	}
    	
    	
    	//requestContext.execute("PF('noWinnersDialog').show()");
    	//requestContext.execute("PF('timerRedirect').start()");
    }
    
    public void winnerRedirect(){  
    	
    	this.setControlledDisconnection(true);
    	//redirect at the end to pictionary page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "pictionary.xhtml?userlogged=" + this.username );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }
    
    public void adivinaRedirect(){  
    	
    	this.setControlledDisconnection(true);
    	//redirect at the end to adivina page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "adivina.xhtml?userlogged=" + this.username );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }  
    
    public void adivinaRedirectWinning(){
    	
    	this.setControlledDisconnection(true);
    	
    	// painter wins 5 points
    	this.scores.increase(users.getPositionOfUser(this.username), 5);
    	
    	//redirect to adivina page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "adivina.xhtml?userlogged=" + this.username );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }
    
    public void adivinaRedirectLosing(){
    	
    	this.setControlledDisconnection(true);

    	// painter loses 5 points
    	this.scores.decrease(users.getPositionOfUser(this.username), 5);
    	
    	//redirect to adivina page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "adivina.xhtml?userlogged=" + this.username );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }  

    // return true if there are only 1 person
    public boolean noMoreUsers() {
    	
    	boolean emptyList = true;
    	
    	if (users.onlyOnePerson())
    		emptyList = true;
    	else
    		emptyList = false;
    	
    	return emptyList;
    }
    
    public void start() {
    	// get the context
    	RequestContext requestContext = RequestContext.getCurrentInstance();

    	//requestContext.update("form:contenido");
    	
    	if (users.onlyOnePerson()) {
    		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "Se necesitan al menos dos jugadores para empezar.", ""));
    	} else {
    		//push start information
    		requestContext.execute("PF('dlg').hide()");
    		eventBus.publish(CHANNEL + "*", new Message(
        		username, String.format(" Comienza la partida!"), true , this.keyword, true));
    		requestContext.execute("PF('timerGame').start()");
    	}
    }
    
    // pass the keyword passed by javascript parameter to this.keyword (adivina site)
    public void passKeyword(){
    	String value = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap().get("param");
    	this.keyword = value;
    	this.controlledDisconnection = false;
    }
    
    // pass the winner passed by javascript parameter to this.winner (adivina site) and show dialog to losers
    public void passWinner(){
    	String value = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap().get("winner");
    	this.winner = value;
        
    	RequestContext requestContext = RequestContext.getCurrentInstance();
    	requestContext.update("form:someDialogHeader");
		requestContext.execute("PF('dialogEnd').show()");
    }

    // get a random word to paint (pictionary site)
    public String getRandomWord() {
    	
    	// give a random number from 0 to palabras.size()-1
    	int valorEntero = (int) Math.floor(Math.random()*(this.palabras.size()));
    	
    	this.keyword = palabras.get(valorEntero);
    	
    	return this.keyword;
    }
     
    public void login() {
    	// get the context
        RequestContext requestContext = RequestContext.getCurrentInstance();
        
        // check if users exists
        if(users.contains(username)) {
            // not login
        	loggedIn = false;
            FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "Nombre existente", "Elige otro nombre."));
            requestContext.update("growl");
        }
        else {
        	// login
            users.add(username);
            scores.add(0);
            // looks for widget client with name subscriber (socket) and create a new connection
            requestContext.execute("PF('subscriber2').connect('/" + username + "')");
            //push join information
            //eventBus.publish(CHANNEL + "*", username + " se unió al canal.");
            requestContext.execute("PF('timerStart').start()");
            loggedIn = true;
        }
    }
    
    
    public void login2() {
    	// get the context
        RequestContext requestContext = RequestContext.getCurrentInstance();
        
        // check if users exists
        if(users.contains(username)) {
            // not login
        	loggedIn = true;
        	// looks for widget client with name subscriber (socket) and create a new connection
            requestContext.execute("PF('subscriber2').connect('/" + username + "')");
            requestContext.update("growl");
            eventBus.publish("/pictionaryroom/*", 
                	new Message(
                		String.format("%s se unió a la partida",  username), true));
            
        }
        else {
        	// login
            users.add(username);
            scores.add(0);
            // looks for widget client with name subscriber (socket) and create a new connection
            requestContext.execute("PF('subscriber2').connect('/" + username + "')");
            //push join information
            //eventBus.publish(CHANNEL + "*", username + " se unió al canal.");
            eventBus.publish("/pictionaryroom/*", 
                	new Message(
                		String.format("%s se unió a la partida",  username), true));
            loggedIn = true;
        }
    }

    public void disconnect() {
    	
    	this.setControlledDisconnection(true);
    	
    	RequestContext requestContext = RequestContext.getCurrentInstance();
    	
        //remove user and update ui
    	scores.remove(users.getPositionOfUser(username));
        users.remove(username);
        
        requestContext.update("form:users");
        
        //push leave information
        eventBus.publish(CHANNEL + "*", new Message(
        		String.format("%s dejó la partida", username), true));
        
        requestContext.execute("PF('subscriber2').disconnect()");
        
        //reset state
        loggedIn = false;
        username = null;
        
        //redirect at the end to login page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "pictionarylogin.xhtml" );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }
    
    public void disconnect2() {
    	
    	this.setControlledDisconnection(true);
    	
RequestContext requestContext = RequestContext.getCurrentInstance();
    	
        //remove user and update ui
		scores.remove(users.getPositionOfUser(username));
        users.remove(username);
        requestContext.update("form:users");
        
        //push leave information
        eventBus.publish(CHANNEL + "*", new Message(
        		String.format("%s dejó la partida", username), true));
        
        // call another player to paint
        if (users.size() >= 1) {
        	eventBus.publish(CHANNEL + users.getUser(0), new Message(true, this.username,
            		String.format("%s dejó la partida", username),true));
        	for (int i = 1;i < users.size();i++) {
        		eventBus.publish(CHANNEL + users.getUser(i), new Message(this.username, true,
                		String.format("%s dejó la partida", username),true));
        	}
        	
        }
        
        requestContext.execute("PF('subscriber2').disconnect()");
        
        //reset state
        loggedIn = false;
        username = null;
        
        //redirect at the end to login page
        try{
        	FacesContext contex = FacesContext.getCurrentInstance();
        	contex.getExternalContext().redirect( "pictionarylogin.xhtml" );
        } catch (Exception e) {
			System.out.println(e.toString());// TODO: handle exception
		}
    }
    
    public void removeUserAdivina() {

    	if (!controlledDisconnection) {
    		RequestContext requestContext = RequestContext.getCurrentInstance();

    		//remove user and update ui
    		scores.remove(users.getPositionOfUser(username));
    		users.remove(username);

    		requestContext.update("form:users");

    		//push leave information
    		eventBus.publish(CHANNEL + "*", new Message(
    				String.format("%s dejó la partida", username), true));

    		requestContext.execute("PF('subscriber2').disconnect()");

    		//reset state
    		loggedIn = false;
    		username = null;
    	}
    }
    
    public void removeUserPictionary() {

    	if (!controlledDisconnection) {
    		RequestContext requestContext = RequestContext.getCurrentInstance();
        	
            //remove user and update ui
    		scores.remove(users.getPositionOfUser(username));
            users.remove(username);
            requestContext.update("form:users");
            
            //push leave information
            eventBus.publish(CHANNEL + "*", new Message(
            		String.format("%s dejó la partida", username), true));
            
            // call another player to paint
            if (users.size() >= 1) {
            	eventBus.publish(CHANNEL + users.getUser(0), new Message(true, this.username,
                		String.format("%s dejó la partida", username),true));
            	for (int i = 1;i < users.size();i++) {
            		eventBus.publish(CHANNEL + users.getUser(i), new Message(this.username, true,
                    		String.format("%s dejó la partida", username),true));
            	}	
            }
     
            requestContext.execute("PF('subscriber2').disconnect()");
            
            //reset state
            loggedIn = false;
            username = null;
    	}
    }

   
    
	public String getColor() {
		return color;
	}

	public void setColor(String color) {
		this.color = color;
	}

	public boolean isControlledDisconnection() {
		return controlledDisconnection;
	}

	public void setControlledDisconnection(boolean controlledDisconnection) {
		this.controlledDisconnection = controlledDisconnection;
	}
}