package es.us.dit.ejemplo.push.chat;

import java.io.Serializable;
import org.primefaces.context.RequestContext;
import org.primefaces.push.EventBus;
import org.primefaces.push.EventBusFactory;

import es.us.dit.ejemplo.pictionary.Message;

import javax.faces.application.FacesMessage;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;
 
@ManagedBean
@ViewScoped
public class ChatView implements Serializable {
     
    //private final PushContext pushContext = PushContextFactory.getDefault().getPushContext();
 
    private final EventBus eventBus = EventBusFactory.getDefault().eventBus();
 
    @ManagedProperty("#{chatUsers}")
    private ChatUsers users;
 
    private String privateMessage;
     
    private String globalMessage;
     
    private String username;
     
    private boolean loggedIn;
     
    private String privateUser;
     
    private final static String CHANNEL = "/chatroom/";
 
    public ChatUsers getUsers() {
        return users;
    }
 
    public void setUsers(ChatUsers users) {
        this.users = users;
    }
     
    public String getPrivateUser() {
        return privateUser;
    }
 
    public void setPrivateUser(String privateUser) {
        this.privateUser = privateUser;
    }
 
    public String getGlobalMessage() {
        return globalMessage;
    }
 
    public void setGlobalMessage(String globalMessage) {
        this.globalMessage = globalMessage;
    }
 
    public String getPrivateMessage() {
        return privateMessage;
    }
 
    public void setPrivateMessage(String privateMessage) {
        this.privateMessage = privateMessage;
    }
     
    public String getUsername() {
        return username;
    }
    public void setUsername(String username) {
        this.username = username;
    }
     
    public boolean isLoggedIn() {
        return loggedIn;
    }
    public void setLoggedIn(boolean loggedIn) {
        this.loggedIn = loggedIn;
    }
 
    public void sendGlobal() {
        eventBus.publish(CHANNEL + "*", username + ": " + globalMessage);
         
        globalMessage = null;
    }
     
    public void sendPrivate() {
        eventBus.publish(CHANNEL + privateUser, "[PM] " + username + ": " + privateMessage);
         
        privateMessage = null;
    }
     
    public void login() {
    	// get the context
        RequestContext requestContext = RequestContext.getCurrentInstance();
        
        // check if users exists
        if(users.contains(username)) {
            // not login
        	loggedIn = false;
            FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(FacesMessage.SEVERITY_ERROR, "Nombre existente", "Elige otro nombre."));
            requestContext.update("growl");
        }
        else {
        	// login
            users.add(username);
            // looks for widget client with name subscriber (socket) and create a new connection
            requestContext.execute("PF('subscriber').connect('/" + username + "')");
            //push join information
            eventBus.publish("/chatroom/*", 
                	new Message(
                		String.format("%s se unió al canal",  username), true));
            loggedIn = true;
        }
    }
     
    public void disconnect() {
    	
    	RequestContext requestContext = RequestContext.getCurrentInstance();
        
        //remove user and update ui
        users.remove(username);
        requestContext.update("form:users");
        
        //push leave information
        eventBus.publish(CHANNEL + "*", new Message(
        		String.format("%s dejó el canal", username), true));
        
        requestContext.execute("PF('subscriber').disconnect()");
        //reset state
        loggedIn = false;
        username = null;
    }
}